@@ -51,17 +51,20 @@
 #include <sys/ioctl.h>
 #include <sys/time.h>
 #include <sys/file.h>
-#include <sys/signal.h>
+#include <signal.h>
+#include <errno.h>
+#include <string.h>
+#include <getopt.h>
+#include <unistd.h>
+
 #include "nls.h"
 
-#ifdef __linux__
-#include <unistd.h>
-#include <string.h>
+#if HAVE_LIBUTIL && HAVE_PTY_H
+#include <pty.h>
 #endif
 
-#include "../defines.h"
-#ifdef HAVE_openpty
-#include <pty.h>
+#ifdef HAVE_LIBUTEMPTER
+#include <utempter.h>
 #endif
 
 void finish(int);
@@ -77,27 +80,32 @@ void doshell(void);
 
 char	*shell;
 FILE	*fscript;
-int	master;
+int	master = -1;
 int	slave;
 int	child;
 int	subchild;
+int	childstatus;
 char	*fname;
 
 struct	termios tt;
 struct	winsize win;
 int	lb;
 int	l;
-#ifndef HAVE_openpty
+#if !HAVE_LIBUTIL || !HAVE_PTY_H
 char	line[] = "/dev/ptyXX";
 #endif
 int	aflg = 0;
 char	*cflg = NULL;
+int	eflg = 0;
 int	fflg = 0;
 int	qflg = 0;
 int	tflg = 0;
 
 static char *progname;
 
+int die;
+int resized;
+
 static void
 die_if_link(char *fn) {
 	struct stat s;
@@ -113,8 +121,19 @@ die_if_link(char *fn) {
 	}
 }
 
+/*
+ * script -t prints time delays as floating point numbers
+ * The example program (scriptreplay) that we provide to handle this
+ * timing output is a perl script, and does not handle numbers in
+ * locale format (not even when "use locale;" is added).
+ * So, since these numbers are not for human consumption, it seems
+ * easiest to set LC_NUMERIC here.
+ */
+
 int
 main(int argc, char **argv) {
+	sigset_t block_mask, unblock_mask;
+	struct sigaction sa;
 	extern int optind;
 	char *p;
 	int ch;
@@ -125,18 +144,19 @@ main(int argc, char **argv) {
 
 
 	setlocale(LC_ALL, "");
+	setlocale(LC_NUMERIC, "C");	/* see comment above */
 	bindtextdomain(PACKAGE, LOCALEDIR);
 	textdomain(PACKAGE);
-	
+
 	if (argc == 2) {
 		if (!strcmp(argv[1], "-V") || !strcmp(argv[1], "--version")) {
-			printf(_("%s from %s\n"),
-			       progname, util_linux_version);
+			printf(_("%s (%s)\n"),
+			       progname, PACKAGE_STRING);
 			return 0;
 		}
 	}
 
-	while ((ch = getopt(argc, argv, "ac:fqt")) != -1)
+	while ((ch = getopt(argc, argv, "ac:efqt")) != -1)
 		switch((char)ch) {
 		case 'a':
 			aflg++;
@@ -144,6 +164,9 @@ main(int argc, char **argv) {
 		case 'c':
 			cflg = optarg;
 			break;
+		case 'e':
+			eflg++;
+			break;
 		case 'f':
 			fflg++;
 			break;
@@ -156,7 +179,7 @@ main(int argc, char **argv) {
 		case '?':
 		default:
 			fprintf(stderr,
-				_("usage: script [-a] [-f] [-q] [-t] [file]\n"));
+				_("usage: script [-a] [-e] [-f] [-q] [-t] [file]\n"));
 			exit(1);
 		}
 	argc -= optind;
@@ -182,14 +205,33 @@ main(int argc, char **argv) {
 		printf(_("Script started, file is %s\n"), fname);
 	fixtty();
 
-	(void) signal(SIGCHLD, finish);
+#ifdef HAVE_LIBUTEMPTER
+	utempter_add_record(master, NULL);
+#endif
+	/* setup SIGCHLD handler */
+	sigemptyset(&sa.sa_mask);
+	sa.sa_flags = 0;
+	sa.sa_handler = finish;
+	sigaction(SIGCHLD, &sa, NULL);
+
+	/* init mask for SIGCHLD */
+	sigprocmask(SIG_SETMASK, NULL, &block_mask);
+	sigaddset(&block_mask, SIGCHLD);
+
+	sigprocmask(SIG_SETMASK, &block_mask, &unblock_mask);
 	child = fork();
+	sigprocmask(SIG_SETMASK, &unblock_mask, NULL);
+
 	if (child < 0) {
 		perror("fork");
 		fail();
 	}
 	if (child == 0) {
+
+		sigprocmask(SIG_SETMASK, &block_mask, NULL);
 		subchild = child = fork();
+		sigprocmask(SIG_SETMASK, &unblock_mask, NULL);
+
 		if (child < 0) {
 			perror("fork");
 			fail();
@@ -198,8 +240,10 @@ main(int argc, char **argv) {
 			dooutput();
 		else
 			doshell();
-	} else
-		(void) signal(SIGWINCH, resize);
+	} else {
+		sa.sa_handler = resize;
+		sigaction(SIGWINCH, &sa, NULL);
+	}
 	doinput();
 
 	return 0;
@@ -212,34 +256,45 @@ doinput() {
 
 	(void) fclose(fscript);
 
-	while ((cc = read(0, ibuf, BUFSIZ)) > 0)
-		(void) write(master, ibuf, cc);
+	while (die == 0) {
+		if ((cc = read(0, ibuf, BUFSIZ)) > 0) {
+			ssize_t wrt = write(master, ibuf, cc);
+			if (wrt == -1) {
+				int err = errno;
+				fprintf (stderr, _("%s: write error %d: %s\n"),
+					progname, err, strerror(err));
+				fail();
+			}
+		}
+		else if (cc == -1 && errno == EINTR && resized)
+			resized = 0;
+		else
+			break;
+	}
+
 	done();
 }
 
 #include <sys/wait.h>
 
 void
-finish(int dummy) {
+finish(int dummy __attribute__ ((__unused__))) {
 	int status;
 	register int pid;
-	register int die = 0;
 
 	while ((pid = wait3(&status, WNOHANG, 0)) > 0)
-		if (pid == child)
+		if (pid == child) {
+			childstatus = status;
 			die = 1;
-
-	if (die)
-		done();
+		}
 }
 
 void
-resize(int dummy) {
+resize(int dummy __attribute__ ((__unused__))) {
+	resized = 1;
 	/* transmit window change information to the child */
 	(void) ioctl(0, TIOCGWINSZ, (char *)&win);
 	(void) ioctl(slave, TIOCSWINSZ, (char *)&win);
-
-	kill(child, SIGWINCH);
 }
 
 /*
@@ -253,36 +308,70 @@ my_strftime(char *buf, size_t len, const char *fmt, const struct tm *tm) {
 
 void
 dooutput() {
-	register int cc;
+	register ssize_t cc;
 	time_t tvec;
 	char obuf[BUFSIZ];
 	struct timeval tv;
 	double oldtime=time(NULL), newtime;
+	int flgs = 0;
+	ssize_t wrt;
+	ssize_t fwrt;
 
 	(void) close(0);
-#ifdef HAVE_openpty
+#ifdef HAVE_LIBUTIL
 	(void) close(slave);
 #endif
 	tvec = time((time_t *)NULL);
 	my_strftime(obuf, sizeof obuf, "%c\n", localtime(&tvec));
 	fprintf(fscript, _("Script started on %s"), obuf);
 
-	for (;;) {
+	do {
+		if (die && flgs == 0) {
+			/* ..child is dead, but it doesn't mean that there is
+			 * nothing in buffers.
+			 */
+			flgs = fcntl(master, F_GETFL, 0);
+			if (fcntl(master, F_SETFL, (flgs | O_NONBLOCK)) == -1)
+				break;
+		}
 		if (tflg)
 			gettimeofday(&tv, NULL);
+
+		errno = 0;
 		cc = read(master, obuf, sizeof (obuf));
+
+		if (die && errno == EINTR && cc <= 0)
+			/* read() has been interrupted by SIGCHLD, try it again
+			 * with O_NONBLOCK
+			 */
+			continue;
 		if (cc <= 0)
 			break;
 		if (tflg) {
 			newtime = tv.tv_sec + (double) tv.tv_usec / 1000000;
-			fprintf(stderr, "%f %i\n", newtime - oldtime, cc);
+			fprintf(stderr, "%f %zd\n", newtime - oldtime, cc);
 			oldtime = newtime;
 		}
-		(void) write(1, obuf, cc);
-		(void) fwrite(obuf, 1, cc, fscript);
+		wrt = write(1, obuf, cc);
+		if (wrt < 0) {
+			int err = errno;
+			fprintf (stderr, _("%s: write error: %s\n"),
+				progname, strerror(err));
+			fail();
+		}
+		fwrt = fwrite(obuf, 1, cc, fscript);
+		if (fwrt < cc) {
+			int err = errno;
+			fprintf (stderr, _("%s: cannot write script file, error: %s\n"),
+				progname, strerror(err));
+			fail();
+		}
 		if (fflg)
 			(void) fflush(fscript);
-	}
+	} while(1);
+
+	if (flgs)
+		fcntl(master, F_SETFL, flgs);
 	done();
 }
 
@@ -293,7 +382,7 @@ doshell() {
 #if 0
 	int t;
 
-	t = open(_PATH_TTY, O_RDWR);
+	t = open(_PATH_DEV_TTY, O_RDWR);
 	if (t >= 0) {
 		(void) ioctl(t, TIOCNOTTY, (char *)0);
 		(void) close(t);
@@ -308,6 +397,8 @@ doshell() {
 	(void) dup2(slave, 2);
 	(void) close(slave);
 
+	master = -1;
+
 	shname = strrchr(shell, '/');
 	if (shname)
 		shname++;
@@ -315,9 +406,9 @@ doshell() {
 		shname = shell;
 
 	if (cflg)
-		execl(shell, shname, "-c", cflg, 0);
+		execl(shell, shname, "-c", cflg, NULL);
 	else
-		execl(shell, shname, "-i", 0);
+		execl(shell, shname, "-i", NULL);
 
 	perror(shell);
 	fail();
@@ -330,7 +421,7 @@ fixtty() {
 	rtt = tt;
 	cfmakeraw(&rtt);
 	rtt.c_lflag &= ~ECHO;
-	(void) tcsetattr(0, TCSAFLUSH, &rtt);
+	(void) tcsetattr(0, TCSANOW, &rtt);
 }
 
 void
@@ -353,17 +444,30 @@ done() {
 		}
 		(void) fclose(fscript);
 		(void) close(master);
+
+		master = -1;
 	} else {
-		(void) tcsetattr(0, TCSAFLUSH, &tt);
+		(void) tcsetattr(0, TCSADRAIN, &tt);
 		if (!qflg)
 			printf(_("Script done, file is %s\n"), fname);
+#ifdef HAVE_LIBUTEMPTER
+		if (master >= 0)
+			utempter_remove_record(master);
+#endif
+	}
+
+	if(eflg) {
+		if (WIFSIGNALED(childstatus))
+			exit(WTERMSIG(childstatus) + 0x80);
+		else
+			exit(WEXITSTATUS(childstatus));
 	}
 	exit(0);
 }
 
 void
 getmaster() {
-#ifdef HAVE_openpty
+#if HAVE_LIBUTIL && HAVE_PTY_H
 	(void) tcgetattr(0, &tt);
 	(void) ioctl(0, TIOCGWINSZ, (char *)&win);
 	if (openpty(&master, &slave, NULL, &tt, &win) < 0) {
@@ -398,24 +502,26 @@ getmaster() {
 					return;
 				}
 				(void) close(master);
+				master = -1;
 			}
 		}
 	}
+	master = -1;
 	fprintf(stderr, _("Out of pty's\n"));
 	fail();
-#endif /* not HAVE_openpty */
+#endif /* not HAVE_LIBUTIL */
 }
 
 void
 getslave() {
-#ifndef HAVE_openpty
+#ifndef HAVE_LIBUTIL
 	line[strlen("/dev/")] = 't';
 	slave = open(line, O_RDWR);
 	if (slave < 0) {
 		perror(line);
 		fail();
 	}
-	(void) tcsetattr(slave, TCSAFLUSH, &tt);
+	(void) tcsetattr(slave, TCSANOW, &tt);
 	(void) ioctl(slave, TIOCSWINSZ, (char *)&win);
 #endif
 	(void) setsid();
