@@ -36,6 +36,8 @@
  *
  * @PTTYPE: partition table type (dos, gpt, etc.).
  *
+ * @PART_ENTRY_SCHEME: partition table type
+ *
  * @PART_ENTRY_NAME: partition name (gpt and mac only)
  *
  * @PART_ENTRY_UUID: partition UUID (gpt only)
@@ -44,6 +46,8 @@
  *
  * @PART_ENTRY_FLAGS: partition flags (e.g. boot_ind) or  attributes (e.g. gpt attributes)
  *
+ * @PART_ENTRY_NUMBER: partition number
+ *
  * Example:
  *
  * <informalexample>
@@ -556,7 +560,7 @@ static int partitions_probe(blkid_probe pr, struct blkid_chain *chn)
 	if (chn->binary)
 		partitions_init_data(pr, chn);
 
-	if (pr->prob_flags & BLKID_PARTS_IGNORE_PT)
+	if (!pr->wipe_size && (pr->prob_flags & BLKID_PARTS_IGNORE_PT))
 		goto details_only;
 
 	DBG(DEBUG_LOWPROBE,
